/*!
        @file    fft.h

        @brief

        @author  Yusuke Namekawa (namekawa)
                 $LastChangedBy: aoyama $

        @date    $LastChangedDate:: 2019-01-21 17:06:23 #$

        @version $LastChangedRevision: 1929 $
*/

#ifndef FFT_INCLUDED
#define FFT_INCLUDED

#ifdef USE_FFTWLIB

#ifdef USE_MPI
#include <fftw3-mpi.h>
#include "Communicator/MPI/communicator_mpi.h"
#else
#include <fftw3.h>
#endif

#ifdef USE_OPENMP
#include "ResourceManager/threadManager_OpenMP.h"
#endif

#include "Field/field.h"
#include "Field/index_lex.h"
#include "Parameters/parameters.h"
#include "Parameters/commonParameters.h"

#include "IO/bridgeIO.h"
using Bridge::vout;

#ifdef USE_FACTORY
#include "factory.h"
#endif


//! Base class for FFT class family.

/*!
    This class implements a base class of Fast Fourier Transformation.
                               [7 Sep 2015 Y.Namekawa]
 */

class FFT
{
 public:
  enum Direction
  {
    FORWARD  = FFTW_FORWARD,
    BACKWARD = FFTW_BACKWARD,
    UNDEF
  };

 public:
  FFT()
    : m_vl(CommonParameters::Vlevel())
  {}

  virtual ~FFT() {}

 private:
  // non-copyable
  FFT(const FFT&);
  FFT& operator=(const FFT&);

 public:
  virtual void fft(Field& field) = 0;  // field is overwritten
  virtual void fft(Field& field_out, const Field& field_in) = 0;
  virtual void fft(Field& field_out, const Field& field_in, const Direction dir) = 0;

  virtual void set_parameters(const Parameters& params)
  {
    // if (params.is_empty()) return;

    std::string vlevel;

    if (params.fetch_string("verbose_level", vlevel)) {
      set_parameter_verboselevel(Bridge::BridgeIO::set_verbose_level(vlevel));
    }
  }

  virtual void set_parameters(const std::string& str_fft_direction) {}
  virtual void set_parameter_verboselevel(const Bridge::VerboseLevel vl) { m_vl = vl; }

 protected:
  Bridge::VerboseLevel m_vl;

#ifdef USE_FACTORY
 public:
  typedef FFT *(*ProductCreator)();
  typedef FactoryTemplate<FFT, ProductCreator> Factory;

  static FFT *New(const IdentifierType& subtype)
  {
    ProductCreator p = Factory::Find(subtype);

    return p ? (*p)() : 0;
  }

#ifdef USE_FACTORY_AUTOREGISTER
#else
  static bool init_factory();
#endif
#endif
};
//- #endif of #ifdef USE_FFTWLIB
#endif
#endif
